from pathlib import Path

import ch5mpy as ch
import rich
import typer

from hdfq.evaluation import eval as hdfq_eval
from hdfq.parser import parse


def run(filter: str, path: Path) -> None:
    tree, mode = parse(filter)

    with ch.options(error_mode="ignore"):
        try:
            h5_object = ch.H5Dict.read(path, mode=mode)

        except IsADirectoryError:
            rich.print(f"[bold red]{path} is not a file[/bold red]")
            raise typer.Exit(code=10)

        except OSError:
            rich.print(f"[bold red]{path} is not a valid hdf5 file[/bold red]")
            raise typer.Exit(code=11)

        hdfq_eval(tree, h5_object)



import re
from collections.abc import Iterable

from hdfq import tokens
from hdfq.syntax import Syntax
from hdfq.tokens import Token


def is_int(string: str) -> bool:
    if string[0] in ("-", "+"):
        return string[1:].isdigit()
    return string.isdigit()


def is_bool(string: str) -> bool:
    return string.lower() in ("true", "false")


def lex(string: str) -> Token:
    if is_int(string):
        return tokens.INT(value=int(string))

    if is_bool(string):
        return tokens.BOOLEAN(value=eval(string.capitalize()))

    match string:
        case Syntax.keys:
            return tokens.KEYS

        case Syntax.attributes:
            return tokens.ATTRIBUTES

        case Syntax.attribute_keys:
            return tokens.ATTRIBUTE_KEYS

        case Syntax.sizes:
            return tokens.SIZES

        case Syntax.del_:
            return tokens.DEL

        case _:
            pass

    if string.capitalize() == Syntax.none:
        return tokens.NONE

    string = string.replace('"', "").replace("'", "")
    if string.isidentifier():
        return tokens.IDENTIFIER(value=string)

    match string:
        case Syntax.dot:
            return tokens.DOT

        case Syntax.comma:
            return tokens.COMMA

        case Syntax.equal:
            return tokens.EQUAL

        case Syntax.octothorpe:
            return tokens.OCTOTHORPE

        case Syntax.left_parenthesis:
            return tokens.LEFT_PARENTHESIS

        case Syntax.right_parenthesis:
            return tokens.RIGHT_PARENTHESIS

        case Syntax.left_bracket:
            return tokens.LEFT_BRACKET

        case Syntax.right_bracket:
            return tokens.RIGHT_BRACKET

        case Syntax.left_angle_bracket:
            return tokens.LEFT_ANGLE_BRACKET

        case Syntax.right_angle_bracket:
            return tokens.RIGHT_ANGLE_BRACKET

        case Syntax.pipe:
            return tokens.PIPE

        case _:
            raise SyntaxError(f"Syntax error at : '{string}'")


def tokenize(string: str) -> Iterable[Token]:
    for s in filter(None, re.split(r"([^a-zA-Z0-9_'\"-])", string.replace(" ", ""))):
        yield lex(s)



import functools
from typing import NamedTuple, override

from hdfq.syntax import Syntax


class Token(NamedTuple):
    kind: Syntax
    value: str | int | float | None = None

    @override
    def __repr__(self) -> str:
        if self.value is None:
            return f"Token<{self.kind}>"
        return f"Token<{self.kind}={self.value}>"

    def short_repr(self) -> str:
        return str(self.kind.value) if self.value is None else str(self.value)


KEYS = Token(Syntax.keys)
ATTRIBUTES = Token(Syntax.attributes)
ATTRIBUTE_KEYS = Token(Syntax.attribute_keys)
SIZES = Token(Syntax.sizes)
DEL = Token(Syntax.del_)
NONE = Token(Syntax.none)

DOT = Token(Syntax.dot)
COMMA = Token(Syntax.comma)
EQUAL = Token(Syntax.equal)
OCTOTHORPE = Token(Syntax.octothorpe)
LEFT_PARENTHESIS = Token(Syntax.left_parenthesis)
RIGHT_PARENTHESIS = Token(Syntax.right_parenthesis)
LEFT_BRACKET = Token(Syntax.left_bracket)
RIGHT_BRACKET = Token(Syntax.right_bracket)
LEFT_ANGLE_BRACKET = Token(Syntax.left_angle_bracket)
RIGHT_ANGLE_BRACKET = Token(Syntax.right_angle_bracket)
PIPE = Token(Syntax.pipe)

BOOLEAN = functools.partial(Token, Syntax.boolean)
INT = functools.partial(Token, Syntax.integer)
IDENTIFIER = functools.partial(Token, Syntax.identifier)


def repr_tokens(tokens: list[Token]) -> str:
    return '"' + "".join(map(lambda x: x.short_repr(), tokens)) + '"'



import ch5mpy as ch
import numpy as np
from tqdm import tqdm


def repair_group(corrupted_file: ch.Group, new_file: ch.Group, verbose: bool, in_RAM: bool) -> None:
    iter_keys = tqdm(corrupted_file.keys(), leave=False) if verbose else corrupted_file.keys()

    for key in iter_keys:
        if isinstance(iter_keys, tqdm):
            iter_keys.set_description(f"Processing key {key}")

        try:
            data = corrupted_file[key]

            if isinstance(data, ch.Dataset):
                if in_RAM:
                    new_file.create_dataset(key, data=np.array(data))
                new_file.create_dataset(key, data=data)

            else:
                new_group = new_file.create_group(key)
                repair_group(data, new_group, verbose, in_RAM)

        except RuntimeError:
            continue



from __future__ import annotations

import io
import re
from itertools import repeat
from typing import TYPE_CHECKING, Any, cast

import ch5mpy as ch
import numpy as np
import rich.box
from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.table import Table
from rich.theme import Theme

if TYPE_CHECKING:
    from hdfq.evaluation import EVAL_OBJECT


class H5Highlighter(RegexHighlighter):
    base_style: str = "h5."
    highlights: list[str] = [
        r"(?P<bool_true>True)",
        r"(?P<bool_false>False)",
        r"(?P<builtin>(?<!\w)(object|None|int\d*|float\d*|[><\|][USB]\d+))",
        r"(?P<number>(?<!\w)[+-]?\d+(?:\.\d*)?(?:e(?:[+-](?:\d+)?)?)?…?(?!\w))",
        r"(?P<str>[b]?[\"'].*?[\"'…])",
        r"(?P<identifier>\.[a-zA-Z_]\w*(?=(?:\[.*\])?:))",
        r"(?P<attribute>#[a-zA-Z_]\w*(?=:))",
        r"(?P<value>\.[a-zA-Z_]\w*?(?==))",
    ]


theme = Theme(
    {
        "h5.bool_true": "bold green",
        "h5.bool_false": "bold red",
        "h5.builtin": "orchid",
        "h5.number": "cyan",
        "h5.str": "green",
        "h5.identifier": "bold yellow",
        "h5.attribute": "italic grey70",
        "h5.value": "blue",
    }
)
console = Console(highlighter=H5Highlighter(), theme=theme)


def get_tabs(offset: int) -> str:
    return "  " * offset


def repr_dict(obj: dict[str, Any] | ch.AttributeManager | ch.H5Dict[Any], offset: int) -> str:
    prefix = "." if isinstance(obj, (dict, ch.H5Dict)) else "#"
    tabs = get_tabs(offset)
    return (
        f"{tabs}"
        + f",\n{tabs}".join(map(lambda kv: f"{prefix}{kv[0]}: {repr_object(kv[1], offset=offset)}", obj.items()))
        + ",\n"
    )


def repr_array_1d(obj: ch.H5Array[Any], table: Table) -> None:
    if len(obj) <= 6:
        table.add_row(*map(repr, obj))

    else:
        table.add_row(*map(repr, obj[:3]), "...", *map(repr, obj[-3:]))


def repr_array_2d(obj: ch.H5Array[Any], table: Table) -> None:
    if obj.shape[0] <= 6:
        for row in range(obj.shape[0]):
            repr_array_1d(obj[row], table)

    else:
        n_cols = obj.shape[1]
        repr_array_1d(obj[0], table)
        repr_array_1d(obj[1], table)
        repr_array_1d(obj[2], table)
        table.add_row(*repeat("...", n_cols if n_cols <= 6 else 7))
        repr_array_1d(obj[obj.shape[0] - 3], table)
        repr_array_1d(obj[obj.shape[0] - 2], table)
        repr_array_1d(obj[obj.shape[0] - 1], table)


def repr_object(obj: EVAL_OBJECT, offset: int) -> str:
    if isinstance(obj, (ch.H5Dict, ch.H5List)):
        if len(obj) + len(obj.attributes) == 0:
            return "{}"

        if len(obj.attributes):
            attributes = repr_dict(obj.attributes, offset=offset + 1)
        else:
            attributes = ""

        if len(obj):
            if isinstance(obj, ch.H5List):
                body = repr_dict(obj.to_dict(), offset=offset + 1)
            else:
                body = repr_dict(obj, offset=offset + 1)
        else:
            body = ""

        tabs = get_tabs(offset)

        return f"{{\n{attributes}{body}{tabs}}}"

    elif isinstance(obj, (dict, ch.AttributeManager)):
        if offset > 0:
            tabs = get_tabs(offset)
            return f"{{\n{repr_dict(obj, offset=offset + 1)}{tabs}}}"

        return f"{{\n{repr_dict(obj, offset=offset + 1)}}}"

    elif isinstance(obj, list):
        tabs = get_tabs(offset + 1)
        content_repr = f",\n{tabs}".join(map(repr, obj))
        return f"[\n{tabs}{content_repr}\n]"

    elif isinstance(obj, ch.H5Array):
        if obj.size == 0:
            return "[]"

        if obj.ndim == 0:
            return str(obj)

        table = Table(show_header=False, show_lines=False, box=rich.box.ROUNDED)
        if obj.ndim == 1:
            repr_array_1d(obj, table)

        elif obj.ndim == 2:
            repr_array_2d(obj, table)

        else:
            table.add_row("...")

        tabs = get_tabs(offset)

        table_console = Console(file=io.StringIO(), width=console.width - 1 - len(tabs))
        table_console.print(f".shape={obj.shape}  .dtype={obj.dtype}", table)
        table_repr = cast(str, table_console.file.getvalue()).rstrip()  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType]
        table_repr = re.sub("\n", f"\n{tabs}", table_repr)

        return table_repr

    elif isinstance(obj, np.void):
        return re.sub(
            r"\.$", "", re.sub(r"\.+", ".", "".join(chr(c) if 32 <= c < 128 else "." for c in bytes(obj)[13:]))
        )

    else:
        return repr(obj)


def display(obj: EVAL_OBJECT) -> None:
    console.print(repr_object(obj, offset=0))


def nice_size_format(size: int) -> str:
    if size < 1e3:
        return f"{size}B"

    elif size < 1e6:
        return f"{size / 1e3:.2f}kB"

    elif size < 1e9:
        return f"{size / 1e6:.2f}MB"

    else:
        return f"{size / 1e9:.2f}GB"



from __future__ import annotations

from enum import Enum


class Syntax(str, Enum):
    # Keywords ------------------------
    keys = "keys"
    attributes = "attrs"
    attribute_keys = "kattrs"
    sizes = "sizes"
    del_ = "del"
    none = "None"

    # Punctuation ---------------------
    dot = "."
    comma = ","
    equal = "="
    octothorpe = "#"
    left_parenthesis = "("
    right_parenthesis = ")"
    left_bracket = "["
    right_bracket = "]"
    left_angle_bracket = "<"
    right_angle_bracket = ">"
    pipe = "|"

    # Literal -------------------------
    boolean = "boolean"
    integer = "int"
    identifier = "identifier"



import sys
from pathlib import Path
from typing import Annotated, cast

import ch5mpy as ch
import click
import typer
from typer.rich_utils import rich_format_error

from hdfq.repair import repair_group

tools = typer.Typer(add_completion=False, pretty_exceptions_enable=False, no_args_is_help=True)


@tools.command(no_args_is_help=True)
def repair(
    ctx: typer.Context,
    path: Annotated[
        Path,
        typer.Argument(
            help="Path to a hdf5 file to repair",
            show_default=False,
        ),
    ] = cast(Path, ... if sys.stdin.isatty() else Path(sys.stdin.read().strip())),
    verbose: Annotated[bool, typer.Option("--verbose", "-v", help="verbose output")] = False,
    in_RAM_copy: Annotated[bool, typer.Option("--in-RAM-copy", "-R", help="perform in-RAM copy of datasets")] = False,
) -> None:
    """
    Repair corrupted HDF5 file by extracting valid groups and datasets.
    """
    if not path.exists():
        rich_format_error(click.UsageError(f"{path} does not exist for 'PATH'.", ctx=ctx))

        raise typer.Exit(code=1)

    restore_path = path.with_stem("~" + path.stem)

    with ch.File(path, mode=ch.H5Mode.READ) as corrupted_file, ch.File(
        restore_path, mode=ch.H5Mode.WRITE_TRUNCATE
    ) as new_file:
        repair_group(corrupted_file, new_file, verbose, in_RAM_copy)

    path.unlink()
    restore_path.rename(path)


if __name__ == "__main__":
    tools()



import sys
from importlib.metadata import version
from pathlib import Path
from typing import Annotated, cast

import click
import typer
from typer.rich_utils import rich_format_error

from hdfq.hdfq import run

app = typer.Typer(add_completion=False, pretty_exceptions_enable=False, no_args_is_help=True)


def version_callback(value: bool):
    if value:
        print(f"hdfq-{version('hdfq')}")
        raise typer.Exit()


@app.command(no_args_is_help=True)
def main(
    ctx: typer.Context,
    filter: Annotated[str, typer.Argument(help="Command filter to evaluate", show_default=False)],
    path: Annotated[
        Path,
        typer.Argument(
            help="Path to a hdf5 file to run the filter on",
            show_default=False,
        ),
    ] = cast(Path, ... if sys.stdin.isatty() else Path(sys.stdin.read().strip())),  # pyright: ignore[reportCallInDefaultInitializer]
    version: Annotated[  # pyright: ignore[reportUnusedParameter]
        bool | None, typer.Option("--version", help="Print current version and quit", callback=version_callback)
    ] = None,
) -> None:
    """
    Command-line processor for viewing and manipulating objects and attributes in a HDF5 file.
    It uses a syntax similar to jq's.
    See https://gitlab.vidium.fr/vidium/hdfq for documentation on supported filters.
    """
    if not path.exists():
        rich_format_error(click.UsageError(f"{path} does not exist for 'PATH'.", ctx=ctx))

        raise typer.Exit(code=1)

    run(filter, path)


if __name__ == "__main__":
    app()



from __future__ import annotations

import functools
import itertools
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Literal, cast, override

import ch5mpy as ch

import hdfq
from hdfq.exceptions import (
    BinaryOpContext,
    ContextInfo,
    DatasetCreationContext,
    FunctionCallContext,
    GetStatementContext,
    ParseError,
)
from hdfq.lexer import tokenize
from hdfq.syntax import Syntax
from hdfq.tokens import Token, repr_tokens, COMMA


class Special(str, Enum):
    context = "CTX"


@dataclass
class Node:
    name: str

    @override
    def __repr__(self) -> str:
        return f"{self.name}()"


@dataclass
class VNode(Node):
    value: str | int | float

    @override
    def __repr__(self) -> str:
        return f"{self.name}(value={self.value})"


@dataclass
class VTNode(Node):
    target: Literal[Special.context] | Node
    value: str | Node

    @override
    def __repr__(self) -> str:
        return f"{self.name}(target={self.target}, value={self.value})"

    def unwrap(self) -> tuple[Node | str, Node]:
        return self.target, VTNode(self.name, Special.context, self.value)


@dataclass
class DatasetNode(Node):
    data: list[Node] | None
    shape: tuple[int, ...] | None
    dtype: str | None
    chunks: bool
    maxshape: bool | tuple[int | None, ...] | None


class Nodes(functools.partial[Node], Enum):
    Display = functools.partial(Node, "Display")
    Keys = functools.partial(Node, "Keys")
    Attrs = functools.partial(Node, "Attrs")
    AttrKeys = functools.partial(Node, "AttrKeys")
    Sizes = functools.partial(Node, "Size")
    Constant = functools.partial(VNode, "Constant")
    Get = functools.partial(VTNode, name="Get")
    GetAttr = functools.partial(VTNode, name="GetAttr")
    Assign = functools.partial(VTNode, name="Assign")
    Del = functools.partial(VTNode, name="Del")
    Dataset = functools.partial(DatasetNode, name="Dataset")


class Tree:
    def __init__(self, body: list[Node]) -> None:
        self.body: list[Node] = body

    @override
    def __repr__(self) -> str:
        return f"AST(\n\tbody={self.body}\n)"


def _all_int(tokens: list[Token]) -> bool:
    return bool(len(tokens)) and all(t.kind == Syntax.integer for t in tokens)


def match_atom(tokens: list[Token]) -> Node | None:
    match tokens:
        # match integer or str identifier
        case [Token(Syntax.integer, value=value)] | [Token(Syntax.identifier, value=value)]:
            return Nodes.Constant(value=value)

        # match float of type ".123"
        # case [hdfq.tokens.DOT, *right] if _all_int(right):
        #     return Nodes.Constant(value=float("." + "".join(str(t.value) for t in right)))

        # match float of type "123."
        case [*left, hdfq.tokens.DOT] if _all_int(left):
            return Nodes.Constant(value=float("".join(str(t.value) for t in left)))

        case _:
            pass

    # match float of type "123.456"
    try:
        index = tokens.index(hdfq.tokens.DOT)
    except ValueError:
        return None

    left, right = tokens[:index], tokens[index + 1 :]
    if not _all_int(left) or not _all_int(right):
        return None

    return Nodes.Constant(value=float("".join(str(t.value) for t in left) + "." + "".join(str(t.value) for t in right)))


def match_multiple(
    tokens: list[Token], *matcher: Callable[[list[Token]], Node | None], sep: Token
) -> list[Node] | None:
    nodes: list[Node] = []

    # split tokens by groups in between occurences of `sep`
    for group in [list(group) for key, group in itertools.groupby(tokens, lambda t: t != sep) if key]:
        for m in matcher:
            node = m(group)
            if node is not None:
                nodes.append(node)
                break

        else:
            return None

    return nodes


def pairwise(iterable: Iterable[Any]) -> Iterator[tuple[Any, Any]]:
    iterator = iter(iterable)
    first, second = None, next(iterator)

    for e in iterator:
        first = second
        second = e
        yield first, second


def split_at_identifiers(tokens: list[Token]) -> list[list[Token]]:
    if tokens[-1] != hdfq.tokens.COMMA:
        tokens.append(hdfq.tokens.COMMA)

    key = itertools.accumulate(
        [a == hdfq.tokens.COMMA and b.kind in Syntax.identifier for a, b in pairwise(tokens)], initial=False
    )
    return [list(group) for _, group in itertools.groupby(tokens, lambda _: next(key))]


def split_after_right_brackets(tokens: list[Token]) -> list[list[Token]]:
    return [
        list(group)
        for test, group in itertools.groupby(
            tokens, lambda t: t in (hdfq.tokens.RIGHT_PARENTHESIS, hdfq.tokens.RIGHT_ANGLE_BRACKET)
        )
        if not test
    ]


def match_shape(shape: tuple[int, ...] | None, tokens: list[Token], allow_none: bool = False) -> tuple[int, ...] | None:
    match tokens:
        case [hdfq.tokens.LEFT_PARENTHESIS, *shape_]:
            is_comma = [(s == hdfq.tokens.COMMA) if i % 2 else False for i, s in enumerate(shape_)]
            is_integer = [
                (isinstance(s.value, int) or allow_none and s.value is None) if not i % 2 else False
                for i, s in enumerate(shape_)
            ]

            if not all(c or i for c, i in zip(is_comma, is_integer)):
                raise ParseError(f"Got unexpected pattern ({repr_tokens(shape_)}) while trying to match a shape")

            if shape is not None:
                raise ParseError(f"Redefinition of dataset shape ({repr_tokens(shape_)}), previous shape was {shape}")

            return tuple(cast(int, s.value) for i, s in enumerate(shape_) if not i % 2)

        case _:
            return shape


def match_dtype(dtype: str | None, tokens: list[Token]) -> str | None:
    match tokens:
        case [hdfq.tokens.LEFT_ANGLE_BRACKET, Token(Syntax.identifier, value=value)]:
            if not isinstance(value, str):
                raise ParseError(f"Got unexpected value '{value}' for data type", context=DatasetCreationContext())

            if dtype is not None:
                raise ParseError(f"Redefinition of data type {value}, previous shape was {dtype}")

            return value

        case _:
            return dtype


def match_dataset(tokens: list[Token]) -> DatasetNode | None:
    try:
        rb_index = tokens.index(hdfq.tokens.RIGHT_BRACKET)
    except ValueError:
        return None

    data, parameters = tokens[:rb_index], tokens[rb_index + 1 :]

    match data:
        case [hdfq.tokens.LEFT_BRACKET]:
            dataset = cast(DatasetNode, Nodes.Dataset(data=None, shape=None, dtype=None, chunks=True, maxshape=None))

        case [hdfq.tokens.LEFT_BRACKET, *content]:
            data_part, *details = split_at_identifiers(content)

            data = match_multiple(
                data_part[:-1], match_atom, match_get_statement, sep=COMMA
            )  # :-1 because of trailing commas
            if data is None:
                raise ParseError(f"Got unexpected pattern {repr_tokens(data_part)}", context=DatasetCreationContext())

            dataset = cast(DatasetNode, Nodes.Dataset(data=data, shape=None, dtype=None, chunks=True, maxshape=None))

            for detail in details:
                match detail:
                    case [
                        Token(Syntax.identifier, value="chunks"),
                        hdfq.tokens.EQUAL,
                        Token(Syntax.boolean, value=boolean),
                        hdfq.tokens.COMMA,
                    ]:
                        dataset.chunks = cast(bool, boolean)

                    case [
                        Token(Syntax.identifier, value="maxshape"),
                        hdfq.tokens.EQUAL,
                        Token(Syntax.boolean, value=maxshape),
                        hdfq.tokens.RIGHT_PARENTHESIS,
                        hdfq.tokens.COMMA,
                    ]:
                        assert isinstance(maxshape, bool)
                        dataset.maxshape = maxshape

                    case [
                        Token(Syntax.identifier, value="maxshape"),
                        hdfq.tokens.EQUAL,
                        *shape,
                        hdfq.tokens.RIGHT_PARENTHESIS,
                        hdfq.tokens.COMMA,
                    ]:
                        maxshape = match_shape(None, shape, allow_none=True)
                        if maxshape is None:
                            raise ParseError("TODO")

                        dataset.maxshape = maxshape

                    case _:
                        raise ParseError(f"Got unexpected pattern '{repr_tokens(detail)}' for dataset creation")

        case _:
            return None

    shape, dtype = None, None
    for parameter in split_after_right_brackets(parameters):
        shape = match_shape(shape, parameter)
        dtype = match_dtype(dtype, parameter)

    dataset.shape = shape
    dataset.dtype = dtype
    return dataset


def matches_whole(tokens: list[Token], allow_empty: bool) -> bool:
    match tokens:
        case []:
            return allow_empty

        case [hdfq.tokens.DOT]:
            return True

        case _:
            return False


def match_get_object(tokens: list[Token], *, allow_get_attr: bool, context: ContextInfo | None) -> VTNode:
    match tokens:
        case [*left, hdfq.tokens.DOT, Token(Syntax.identifier | Syntax.integer, value=value)]:
            target = match_get_statement(left, context=context) if len(left) else Special.context
            return cast(VTNode, Nodes.Get(target=target, value=value))

        case [*left, hdfq.tokens.OCTOTHORPE, Token(Syntax.identifier, value=value)]:
            if not allow_get_attr:
                if isinstance(context, GetStatementContext):
                    context.first = value
                raise ParseError("Cannot get attribute ", context=context)

            if context is None:
                context = GetStatementContext(second=str(value))
            target = match_get_statement(left, allow_get_attr=False, context=context) if len(left) else Special.context
            return cast(VTNode, Nodes.GetAttr(target=target, value=value))

        case _:
            raise ParseError(f"Got unexpected pattern {repr_tokens(tokens)}", context=context)


def match_get_statement(
    tokens: list[Token], *, allow_get_attr: bool = True, context: ContextInfo | None = None
) -> VTNode:
    return match_get_object(tokens, allow_get_attr=allow_get_attr, context=context)


def match_get_statement_all(
    tokens: list[Token], *, allow_empty: bool = False, context: BinaryOpContext | None = None
) -> Node | None:
    if matches_whole(tokens, allow_empty=allow_empty):
        return None

    return match_get_statement(tokens, context=context)


def match_assignment(tokens: list[Token]) -> Node | None:
    try:
        assign_index = tokens.index(hdfq.tokens.EQUAL)
    except ValueError:
        return None

    left, right = tokens[:assign_index], tokens[assign_index + 1 :]
    return Nodes.Assign(
        target=match_get_statement_all(left, context=BinaryOpContext("assignment", "left")),
        value=match_atom(right)
        or match_dataset(right)
        or match_get_statement_all(right, context=BinaryOpContext("assignment", "right")),
    )


def match_descriptor(tokens: list[Token]) -> Node | None:
    match tokens:
        case [hdfq.tokens.KEYS]:
            return Nodes.Keys()

        case [hdfq.tokens.ATTRIBUTES]:
            return Nodes.Attrs()

        case [hdfq.tokens.ATTRIBUTE_KEYS]:
            return Nodes.AttrKeys()

        case [hdfq.tokens.SIZES]:
            return Nodes.Sizes()

        case _:
            return None


def match_function_call(tokens: list[Token]) -> Node | None:
    match tokens:
        case [hdfq.tokens.DEL, hdfq.tokens.LEFT_PARENTHESIS, *argument, hdfq.tokens.RIGHT_PARENTHESIS]:
            target, value = match_get_statement(argument, context=FunctionCallContext("del")).unwrap()
            return Nodes.Del(target=target, value=value)

        case _:
            return None


def match_statement(tokens: list[Token]) -> tuple[Node | None, bool]:
    node = match_assignment(tokens) or match_function_call(tokens)
    if node is not None:
        return node, True

    return (match_descriptor(tokens) or match_get_statement_all(tokens, allow_empty=True)), False


def split_at_pipes(tokens: list[Token]) -> Iterator[list[Token]]:
    try:
        index = tokens.index(hdfq.tokens.PIPE)
        yield tokens[:index]
        yield from split_at_pipes(tokens[index + 1 :])

    except ValueError:
        yield tokens


def match_statements(tokens: list[Token]) -> tuple[list[Node], bool]:
    statements: list[Node] = []
    requires_write_access = False

    for statement in split_at_pipes(tokens):
        matched_statement, is_write_operation = match_statement(statement)
        if matched_statement is not None:
            statements.append(matched_statement)
            requires_write_access = requires_write_access or is_write_operation

    return statements, requires_write_access


def parse(filter: str) -> tuple[Tree, ch.H5Mode]:
    tokens = list(tokenize(filter))
    nodes, requires_write_access = match_statements(tokens)

    return Tree(body=nodes + [Nodes.Display()]), ch.H5Mode.READ_WRITE if requires_write_access else ch.H5Mode.READ



from __future__ import annotations

import functools
import sys
from dataclasses import dataclass
from typing import Any, Literal, Protocol

import ch5mpy as ch
import numpy as np
import numpy.typing as npt

from hdfq.display import display, nice_size_format
from hdfq.exceptions import EvalError
from hdfq.parser import DatasetNode, Node, Special, Tree, VNode, VTNode


@dataclass
class DatasetInfo:
    shape: tuple[int, ...] | None
    dtype: str
    data: ch.Dataset[Any] | npt.NDArray[Any] | None
    chunks: bool
    maxshape: tuple[int | None, ...] | bool


EVAL_OBJECT = (
    ch.H5Dict[Any]
    | ch.Dataset[Any]
    | ch.AttributeManager
    | ch.H5List[ch.H5Dict[Any] | ch.Dataset[Any]]
    | list[str]
    | dict[str, Any]
    | DatasetInfo
    | str
    | int
    | float
)


class EVAL_FUNC_base(Protocol):
    def __call__(self, obj: EVAL_OBJECT) -> EVAL_OBJECT: ...


class EVAL_FUNC_get(Protocol):
    def __call__(self, obj: EVAL_OBJECT, key: str) -> EVAL_OBJECT: ...


class EVAL_FUNC_set(Protocol):
    def __call__(self, obj: EVAL_OBJECT, key: str, value: Any) -> EVAL_OBJECT: ...


EVAL_FUNC = EVAL_FUNC_base | EVAL_FUNC_get | EVAL_FUNC_set


def get_object(obj: EVAL_OBJECT, key: str) -> EVAL_OBJECT:
    if not isinstance(obj, (ch.H5Dict, dict, ch.H5List)):
        raise EvalError(f"Cannot get object from '{type(obj).__name__}'")

    return obj[key]


def get_attribute(obj: EVAL_OBJECT, key: str) -> EVAL_OBJECT:
    if not isinstance(obj, (ch.H5Dict, ch.Dataset)):
        raise EvalError(f"Cannot get attribute from '{type(obj).__name__}'")

    return obj.attributes[key]


def get_keys(obj: EVAL_OBJECT) -> list[str]:
    if isinstance(obj, ch.H5Dict):
        return list(obj.keys())

    if isinstance(obj, ch.H5List):
        return [str(i) for i in range(len(obj))]

    raise EvalError(f"Cannot get keys from '{type(obj).__name__}'")


def get_attributes(obj: EVAL_OBJECT) -> dict[str, Any]:
    if not isinstance(obj, (ch.H5Dict, ch.Dataset)):
        raise EvalError(f"Cannot get attributes from '{type(obj).__name__}'")

    return obj.attributes.as_dict()


def get_attribute_keys(obj: EVAL_OBJECT) -> list[str]:
    if not isinstance(obj, (ch.H5Dict, ch.Dataset)):
        raise EvalError(f"Cannot get attribute keys from '{type(obj).__name__}")

    return list(obj.attributes.keys())


def get_sizes_core(obj: ch.H5Dict, sizes: dict[str, Any], cum_size: int) -> tuple[int, dict[str, Any]]:
    for k, v in obj.items():
        if not isinstance(v, ch.H5Dict):
            if isinstance(v, str):
                v_size = sys.getsizeof(v)

            else:
                v_size = v.size * v.dtype.itemsize

            sizes[k] = nice_size_format(v_size)
            cum_size += v_size

        else:
            sub_cum_size, sub_sizes = get_sizes_core(v, {}, 0)
            sizes[k + f"['{nice_size_format(sub_cum_size)}']"] = sub_sizes
            cum_size += sub_cum_size

    return cum_size, sizes


def get_sizes(obj: EVAL_OBJECT) -> dict[str, Any]:
    if not isinstance(obj, ch.H5Dict):
        raise EvalError(f"Cannot get object sizes from '{type(obj).__name__}")

    cum_sum, sizes = get_sizes_core(obj, {}, 0)
    sizes["TOTAL"] = nice_size_format(cum_sum)

    return sizes


def set_key_value(obj: EVAL_OBJECT, key: str | int, value: Any) -> None:
    if not isinstance(obj, (ch.H5Dict, dict, ch.AttributeManager, DatasetInfo)):
        raise EvalError(f"Cannot assign value to '{type(obj).__name__}'")

    obj[key] = value


def del_object(obj: EVAL_OBJECT, key: str | int) -> None:
    if not isinstance(obj, (ch.H5Dict, dict, ch.AttributeManager, ch.H5List)):
        raise EvalError(f"Cannot delete value from '{type(obj).__name__}'")

    del obj[key]


def shallow_eval_statement(target: VTNode, context: EVAL_OBJECT) -> tuple[EVAL_OBJECT, str | int]:
    context, key = eval_statement(target.target, context), target.value
    assert isinstance(key, str | int), "invalid key type"

    if target.name == "GetAttr":
        if not isinstance(context, ch.H5Dict):
            raise EvalError(f"Cannot get attribute '{key}' from {type(context).__name__}")
        context = context.attributes

    return context, key


def create_dataset(
    data: list[EVAL_OBJECT] | None,
    shape: tuple[int, ...] | None,
    dtype: str | None,
    chunks: bool,
    maxshape: bool | tuple[int | None, ...] | None,
) -> ch.AnonymousArrayCreationFunc:
    if shape is None:
        if data is None:
            shape = (0,)

        else:
            shape = (len(data),)

    match data:
        case None:
            dtype = dtype or "f"
            return ch.empty.defer(shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape)

        case [0]:
            dtype = dtype or "f"
            return ch.zeros.defer(shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape)

        case [1]:
            dtype = dtype or "f"
            return ch.ones.defer(shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape)

        case [int(value)]:
            dtype = dtype or "i"
            return ch.full.defer(shape=shape, fill_value=value, dtype=dtype, chunks=chunks, maxshape=maxshape)

        case [*values] if all(isinstance(value, (int, float)) for value in values):
            dtype = dtype or ("i" if all(isinstance(value, int) for value in values) else "f")
            return functools.partial(
                ch.store_dataset,
                array=np.array(values),
                shape=shape,
                dtype=dtype,
                chunks=chunks,
                maxshape=maxshape,
            )

        case _:
            raise EvalError("TODO", data)


def eval_statement(statement: Node | Literal[Special.context], context: EVAL_OBJECT) -> EVAL_OBJECT:
    match statement:
        case Node(name="Display"):
            display(context)

        case Node(name="Keys"):
            context = get_keys(context)

        case Node(name="Attrs"):
            context = get_attributes(context)

        case Node(name="AttrKeys"):
            context = get_attribute_keys(context)

        case Node(name="Size"):
            context = get_sizes(context)

        case VTNode(name="Get", target=target, value=value):
            context = get_object(eval_statement(target, context), value)

        case VTNode(name="GetAttr", target=target, value=value):
            context = get_attribute(eval_statement(target, context), value)

        case VTNode(name="Assign", target=target, value=value):
            value = eval_statement(value, context)
            context, key = shallow_eval_statement(target, context)
            set_key_value(context, key, value)

        case VTNode(name="Del", target=target, value=value):
            context = eval_statement(target, context)
            context, value = shallow_eval_statement(value, context)
            del_object(context, value)

        case VNode(name="Constant", value=value):
            context = value

        case DatasetNode(name="Dataset", data=data, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape):
            if data is None:
                context = create_dataset(None, shape, dtype, chunks, maxshape)

            else:
                context = create_dataset(
                    [eval_statement(node, context) for node in data], shape, dtype, chunks, maxshape
                )

        case Special.context:
            pass

        case _:
            raise EvalError(f"Got unexpected statement '{statement}'")

    return context


def eval(tree: Tree, context: EVAL_OBJECT) -> None:
    for statement in tree.body:
        context = eval_statement(statement, context)



import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal, override


def myexcepthook(type: type[Exception], value: Exception, _):
    print(f"hdfq: {type.__name__}: {value}")


sys.excepthook = myexcepthook


@dataclass
class ContextInfo(ABC):
    @override
    @abstractmethod
    def __repr__(self) -> str:
        pass


@dataclass
class GetStatementContext(ContextInfo):
    second: str
    _first: str | None = None

    @override
    def __repr__(self) -> str:
        assert self.first is not None
        return f'"{self.second}" of attribute "{self.first}"'

    @property
    def first(self) -> str | None:
        return self._first

    @first.setter
    def first(self, value: str):
        self._first = value


@dataclass
class BinaryOpContext(ContextInfo):
    kind: Literal["assignment"]
    side: Literal["left", "right"]

    @override
    def __repr__(self) -> str:
        return f" while parsing {self.side} hand side of {self.kind}"


@dataclass
class FunctionCallContext(ContextInfo):
    kind: Literal["del"]

    @override
    def __repr__(self) -> str:
        return f" while parsing arguments of {self.kind} function"


@dataclass
class DatasetCreationContext(ContextInfo):
    @override
    def __repr__(self) -> str:
        return " while parsing arguments for dataset creation"


class ParseError(Exception):
    def __init__(self, msg: str = "", context: ContextInfo | None = None) -> None:
        super().__init__(msg)
        self.msg: str = msg
        self.context: ContextInfo | None = context

    def _str_context(self) -> str:
        return "" if self.context is None else repr(self.context)

    @override
    def __str__(self) -> str:
        return f"{self.msg}{self._str_context()}"


class EvalError(Exception):
    pass
